/***************************************************************************
 *   Copyright (C) 2007 by Antony Martin                                   *
 *   martin.antony@yahoo.fr                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/* Cree le : 6 mars 2006
   derniere modification le 1/05/2007 */


#ifndef __SCESHADERS_H__
#define __SCESHADERS_H__


/* constantes propres a Shaders */
#define SCE_UNKNOW_SHADER -1
#define SCE_GLSL_SHADER 1
#define SCE_ASM_SHADER 2
#define SCE_CG_SHADER 3

typedef struct sce_sshader SCE_SShader;
struct sce_sshader
{
    /* shaders, tableaux de 2 pour vertex/pixel shader */
    SCE_CShaderGLSL *s_glsl[2];
    SCE_CShaderASM *s_asm[2];
    #ifdef _SCE_USE_CG
    SCE_CShaderCG *s_cg[2];
    #endif
    /* program GLSL */
    SCE_CProgram *p_glsl;
    
    /* type du shader (langage; GLSL, ASM, Cg) */
    int type;
    
    /* sources */
    char *vs_source, *ps_source;
    
    /* code source additif */
    char *vs_addsrc, *ps_addsrc;
    
    /* booleen, true: le shader est pret pour utilisation */
    int ready;
};



/* initialise le gestionnaire de shaders */
int SCE_Shader_Init(void);

/* quitte le gestionnaire de shaders */
int SCE_Shader_Quit(void);

/* retourne l'identifiant de type du media manager */
int SCE_Shader_GetMediaTypeID(void);


/* cree un shader */
SCE_SShader* SCE_Shader_Create(int);

/* detruit un shader */
void SCE_Shader_Delete(SCE_SShader*);

/* retourne le langage du shader (Cg, GLSL ou ASM) */
int SCE_Shader_GetLanguage(SCE_SShader*);

/* retourne le type du shader (vertex, pixel, ou les deux) */
int SCE_Shader_GetType(SCE_SShader*);



/* charge (char**) un fichier contenant le code du vertex shader
   et/ou le code du pixel shader (prefixes : [Vertex Shader] [Pixel Shader]) */
void* SCE_Shader_LoadSourceFromFile(FILE*, const char*, void*);

/* cree un shader a partir de deux fichiers, le premier contient le code du
   vertex shader, le second le code du pixel shader. un des deux fichiers peut
   contenir les deux types de code ensemble */
SCE_SShader* SCE_Shader_Load(const char*, const char*);

/* construit un shader */
int SCE_Shader_Build(SCE_SShader*);

/* ajoute un morceau de code source qui sera ajoute
   au debut du code source par defaut */
int SCE_Shader_AddSource(SCE_SShader*, int, char*);


/* retourne l'index d'une variable de shader */
int SCE_Shader_GetIndex(SCE_SShader*, int, const char*);
/*
int SCE_Shader_GetParam(SCE_SShader*, const char*);
float SCE_Shader_GetParamf(SCE_SShader*, const char*);
float* SCE_Shader_GetParamfv(SCE_SShader*, const char*);
*/

void SCE_Shader_Param(SCE_SShader*, int, const char*, int);
void SCE_Shader_Paramf(SCE_SShader*, int, const char*, float);
void SCE_Shader_Paramfv(SCE_SShader*, int, const char*, size_t, float*);

void SCE_Shader_SetParam(SCE_SShader*, int, int);
void SCE_Shader_SetParamf(SCE_SShader*, int, float);
void SCE_Shader_SetParamfv(SCE_SShader*, int, size_t, float*);


/* defini le shader comme etant actif (fonction generique) */
void SCE_Shader_Use(SCE_SShader*);

/*
void SCE_Shader_UsePixelShader(SCE_SShader*);

void SCE_Shader_UseVertexShader(SCE_SShader*);
*/


#endif /* guard */
