/***************************************************************************
 *   Copyright (C) 2007 by Antony Martin                                   *
 *   martin.antony@yahoo.fr                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/* Cree le : 6 mars 2006
   derniere modification le 01/05/2007 */


#include <SCE/SCEMinimal.h>

#include <SCE/utils/SCEString.h>

#include <SCE/core/SCECShader.h>

#include <SCE/interface/SCEMedia.h>

#include <SCE/interface/SCEShaders.h>



/* identifiant de type du media manager */
static int shader_mediatypeid;



int SCE_Shader_Init(void)
{
    static int is_init = 0;
    
    if(!is_init)
    {
        shader_mediatypeid = SCE_Media_GenTypeID();
        SCE_Media_RegisterLoader(shader_mediatypeid, 0,
                                 ".glsl .vert .frag .asm .vs .ps .fs"
                                 " .cg .vcg .pcg .fcg",
                                 SCE_Shader_LoadSourceFromFile);
        
        is_init = 1;
    }
    
    return (SCE_OK);
}

int SCE_Shader_Quit(void)
{
    return (SCE_OK);
}

int SCE_Shader_MediaTypeID(void)
{
    return (shader_mediatypeid);
}


SCE_SShader* SCE_Shader_Create(int type)
{
    SCE_SShader *shader = NULL;
    
    #ifdef __SCE_DEBUG__
    if(type != SCE_UNKNOW_SHADER &&
       type != SCE_GLSL_SHADER &&
       type != SCE_ASM_SHADER
       #ifdef _SCE_USE_CG
       && type != SCE_CG_SHADER
       #endif
       )
    {
        Logger_Log(SCE_INVALID_ARG);
        Logger_LogMsg("shader type %d is not reconized", type);
        return (NULL);
    }
    #endif
    
    shader = SCE_malloc(sizeof *shader);
    if(shader == NULL)
    {
        Logger_LogSrc();
        return (NULL);
    }
    
    shader->type = type;
    shader->ready = 0;
    shader->s_glsl[0] = shader->s_glsl[1] = NULL;
    shader->s_asm[0] = shader->s_asm[1] = NULL;
    #ifdef _SCE_USE_CG
    shader->s_cg[0] = shader->s_cg[1] = NULL;
    #endif
    shader->p_glsl = NULL;
    shader->vs_source = shader->ps_source = NULL;
    shader->vs_addsrc = shader->ps_addsrc = NULL;
    
    return (shader);
}

void SCE_Shader_Delete(SCE_SShader *shader)
{
    unsigned int i;
    
    if(shader != NULL)
    {
        SCE_CDeleteProgram(shader->p_glsl);
        shader->p_glsl = NULL;
        
        for(i=0; i<2; i++)
        {
            SCE_CDeleteShaderGLSL(shader->s_glsl[i]);
            shader->s_glsl[i] = NULL;
            SCE_CDeleteShaderASM(shader->s_asm[i]);
            shader->s_asm[i] = NULL;
            #ifdef _SCE_USE_CG
            SCE_CDeleteShaderCG(shader->s_cg[i]);
            shader->s_cg[i] = NULL;
            #endif
        }
        
        SCE_free(shader->vs_addsrc);
        SCE_free(shader->ps_addsrc);
        
        SCE_free(shader->vs_source);
        SCE_free(shader->ps_source);
        
        /* au cas ou ... */
        shader->ready = 0;
        
        SCE_free(shader);
    }
}

int SCE_Shader_GetLanguage(SCE_SShader *shader)
{
    return (shader->type);
}

int SCE_Shader_GetType(SCE_SShader *shader)
{
    if((shader->ps_source != NULL || shader->ps_addsrc != NULL) &&
       (shader->vs_source != NULL || shader->vs_addsrc != NULL) )
        return (SCE_UNKNOW_SHADER);
    else if(shader->ps_source != NULL || shader->ps_addsrc != NULL)
        return (SCE_PIXEL_SHADER);
    else
        return (SCE_VERTEX_SHADER);
}


/* retourne le type d'un shader d'apres son extension */
static void SCE_Shader_SearchTypes(const char *ext, int *type)
{
    #define SCE_SHADER_FOR(str, val, val2)\
    if(SCE_String_Cmp(str, ext, 0))\
    {\
        type[0] = val;\
        type[1] = val2;\
    }\
    else 
    
    /* NOTE: c'est bien statique tout ça... il faudrait rajouter une
             fonctionnalite de stockage d'extensions */
    SCE_SHADER_FOR(".glsl", SCE_GLSL_SHADER, SCE_UNKNOW_SHADER)
    SCE_SHADER_FOR(".vert", SCE_GLSL_SHADER, SCE_VERTEX_SHADER)
    SCE_SHADER_FOR(".frag", SCE_GLSL_SHADER, SCE_PIXEL_SHADER)
    
    SCE_SHADER_FOR(".asm", SCE_ASM_SHADER, SCE_UNKNOW_SHADER)
    SCE_SHADER_FOR(".shd", SCE_ASM_SHADER, SCE_UNKNOW_SHADER)
    SCE_SHADER_FOR(".vs", SCE_ASM_SHADER, SCE_VERTEX_SHADER)
    SCE_SHADER_FOR(".ps", SCE_ASM_SHADER, SCE_PIXEL_SHADER)
    SCE_SHADER_FOR(".fs", SCE_ASM_SHADER, SCE_PIXEL_SHADER)
    SCE_SHADER_FOR(".vsh", SCE_ASM_SHADER, SCE_VERTEX_SHADER)
    SCE_SHADER_FOR(".psh", SCE_ASM_SHADER, SCE_PIXEL_SHADER)
    SCE_SHADER_FOR(".fsh", SCE_ASM_SHADER, SCE_PIXEL_SHADER)
    
    #ifdef _SCE_USE_CG
    SCE_SHADER_FOR(".cg", SCE_CG_SHADER, SCE_UNKNOW_SHADER)
    SCE_SHADER_FOR(".vcg", SCE_CG_SHADER, SCE_VERTEX_SHADER)
    SCE_SHADER_FOR(".pcg", SCE_CG_SHADER, SCE_PIXEL_SHADER)
    SCE_SHADER_FOR(".fcg", SCE_CG_SHADER, SCE_PIXEL_SHADER)
    #endif
    {
        /* NOTE: pourquoi ne pas mettre Cg par defaut ?
                 apres tout il est generique...
                 NON: car il va falloir se faire chier avec les bons profiles */
        Logger_Log(SCE_INVALID_ARG);
        Logger_LogMsg("'%s' is not a valid shader source's extension", ext);
        type[0] = type[1] = SCE_UNKNOW_SHADER;
    }
    
    #undef SCE_SHADER_FOR
}


/* NOTE: a deplacer */
/* positionne le curseur de fp sur la prochaine occurrence a str
   retourne 1 si la fonction a trouve une occurrence */
static int SCE_Shader_SetPosFile(FILE *fp, const char *str, int at_end)
{
    size_t i=0;
    long cur=0, curo=0;
    char buf[256];
    size_t lenstr=0;
    
    lenstr = strlen(str);
    curo = ftell(fp);
    
    buf[lenstr] = '\0';
    
    /* NOTE: algorithme a ameliorer...
             creer une fonction qui propose de relever toutes les positions
             de toutes les chaines passees en parametres (params variables) */
    while(1)
    {
        fseek(fp, curo+cur, SEEK_SET);
        cur++;
        
        for(i=0; i<lenstr; i++)
        {
            buf[i] = fgetc(fp);
            
            if(buf[i] == EOF) /* fin du fichier */
            {
                fseek(fp, curo, SEEK_SET);
                return (0);
            }
        }
        
        if(SCE_String_Cmp(str, buf, 0))
        {
            /* on se replace au debut */
            if(!at_end)
                fseek(fp, -lenstr, SEEK_CUR);
            return (1);
        }
    }
    
    fseek(fp, curo, SEEK_SET);
    
    return (0);
}

/* charge du texte a partir d'un fichier, jusqu'a end */
static char* SCE_Shader_LoadSource(FILE *fp, long end)
{
    char *src = NULL;
    long len = ftell(fp);
    long curpos;
    
    /* calcul de la taille total de la chaine */
    if(end == EOF)
    {
        /* on calcul la distance qu'il y a jusqu'a la fin du fichier */
        curpos = len;
        fseek(fp, 0, SEEK_END);
        len = ftell(fp) - len;
        fseek(fp, curpos, SEEK_SET);
    }
    else
        len = end - len;
    
    
    /* dans le cas ou la zone serait vide, aucun code de shader du type prefixe
       n'est present a cet endroit, mais pas la peine d'en faire un fromage */
    if(len <= 0)
        return (NULL);
    
    /* allocation */
    src = SCE_malloc(len+1);
    if(src == NULL)
    {
        Logger_LogSrc();
        return (NULL);
    }
    src[len] = '\0';
    
    /* lecture */
    fread(src, sizeof(char), len, fp);
    
    return (src);
}

void* SCE_Shader_LoadSourceFromFile(FILE *fp, const char *filename, void *infos)
{
    static char *srcs[2];
    long vpos, ppos, vend, pend;
    int *type = infos;
    const char *ext = SCE_String_GetExt((char*)filename);
    
    /* inutilise */
    filename = NULL;
    
    /* remise a NULL */
    srcs[0] = srcs[1] = NULL;
    
    /* recherche du type du shader */
    SCE_Shader_SearchTypes(ext, type);
    
    if(type[1] == SCE_UNKNOW_SHADER)
    {
        /* le fichier contient le code du vertex et du pixel shader */
        
        /* recuperation de la position du code du vertex shader */
        SCE_Shader_SetPosFile(fp, "[vertex shader]", SCE_FALSE);
        vpos = ftell(fp);
        vend = vpos + 15;   /* 16 = strlen("[vertex shader]"); */
        rewind(fp);
        
        /* recuperation de la position du code du pixel shader */
        SCE_Shader_SetPosFile(fp, "[pixel shader]", SCE_FALSE);
        ppos = ftell(fp);
        pend = ppos + 14;   /* 15 = strlen("[pixel shader]"); */
        
        /* lecture du code du vertex shader */
        fseek(fp, vend, SEEK_SET);
        srcs[0] = SCE_Shader_LoadSource(fp, (vpos > ppos) ? EOF:ppos);
        
        /* lecture du code du pixel shader */
        fseek(fp, pend, SEEK_SET);
        srcs[1] = SCE_Shader_LoadSource(fp, (ppos > vpos) ? EOF:vpos);
    }
    else
    {
        /* le fichier contient le code de type[1] */
        
        srcs[0] = SCE_Shader_LoadSource(fp, EOF);
        
        if(type[1] == SCE_PIXEL_SHADER)
        {
            srcs[1] = srcs[0];
            srcs[0] = NULL;
        }
    }
    
    return (srcs);
}

SCE_SShader* SCE_Shader_Load(const char *vname, const char *pname)
{
    SCE_SShader *shader = NULL;
    char **srcs = NULL;
    char *vsource = NULL, *psource = NULL;
    int type[2];
    int ttemp = SCE_UNKNOW_SHADER;
    
    
    if(vname != NULL)
    {
        srcs = SCE_Media_LoadFromFile(vname, type, NULL);
        /* srcs ne peut pas etre null, il est une variable
           statique de la fonction de chargement */
        if(srcs[0] != NULL && srcs[1] != NULL)
        {
            /* on vient de charger un fichier contenant les deux codes */
            pname = NULL;
        }
        
        vsource = srcs[0];
        psource = srcs[1];
        ttemp = type[0];
    }
    
    if(pname != NULL)
    {
        srcs = SCE_Media_LoadFromFile(pname, type, NULL);
        if(srcs[0] != NULL && srcs[1] != NULL)
        {
            /* on vient de charger un fichier contenant les deux codes */
            vsource = srcs[0];
        }
        
        psource = srcs[1];
        
        if(ttemp != type[0] && vname != NULL)
        {
            /* le type du pixel shader differe de celui du vertex shader */
            
            Logger_Log(SCE_INVALID_ARG);
            
            SCE_free(vsource);
            SCE_free(psource);
            
            Logger_LogMsg("you can't load a %s vertex shader with"
                          " a %s pixel shader",
                          (ttemp == SCE_GLSL_SHADER) ? "GLSL":
                          (ttemp == SCE_ASM_SHADER)  ? "ASM":"Cg",
                          (type[0] == SCE_GLSL_SHADER) ? "GLSL":
                          (type[0] == SCE_ASM_SHADER)  ? "ASM":"Cg");
            
            return (NULL);
        }
    }
    
    
    /* creation du shader */
    shader = SCE_Shader_Create(type[0]);
    if(shader == NULL)
    {
        SCE_free(vsource);
        SCE_free(psource);
        Logger_LogSrc();
        return (NULL);
    }
    
    
    shader->vs_source = vsource;
    shader->ps_source = psource;
    
    return (shader);
}


int SCE_Shader_Build(SCE_SShader *shader)
{
    /* langage du shader */
    int type;
    
    /* taille de final_src */
    size_t final_size = 1;
    
    /* code source final (concatene avec les 'addsrc' du shader) */
    char *final_src = NULL;
    
    
    #ifdef __SCE_DEBUG__
    if(shader == NULL)
    {
        Logger_Log(SCE_INVALID_ARG);
        return (SCE_ERROR);
    }
    /* si le shader ne contient aucun code source */
    if(shader->vs_source == NULL && shader->ps_source == NULL &&
       shader->vs_addsrc == NULL && shader->ps_addsrc == NULL)
    {
        Logger_Log(SCE_INVALID_ARG);
        return (SCE_ERROR);
    }
    #endif
    
    /* recuperation du langage (GLSL, ASM, Cg) du shader */
    type = SCE_Shader_GetLanguage(shader);
    
    #ifdef __SCE_DEBUG__
    if(type == SCE_UNKNOW_SHADER)
    {
        /* type de shader inconnu, arret */
        Logger_Log(SCE_INVALID_ARG);
        return (SCE_ERROR);
    }
    else
    #endif
    if(type == SCE_GLSL_SHADER)
    {
        /* creation des objets */
        /* vertex shader */
        if(shader->vs_source != NULL || shader->vs_addsrc != NULL)
        {
            /* des sources sont disponibles, creation de l'objet */
            shader->s_glsl[0] = SCE_CCreateShaderGLSL(SCE_VERTEX_SHADER);
            if(shader->s_glsl[0] == NULL)
            {
                Logger_LogSrc();
                return (SCE_ERROR);
            }
        }
        
        /* pixel shader */
        if(shader->ps_source != NULL || shader->ps_addsrc != NULL)
        {
            /* des sources sont disponibles, creation de l'objet */
            shader->s_glsl[1] = SCE_CCreateShaderGLSL(SCE_PIXEL_SHADER);
            if(shader->s_glsl[1] == NULL)
            {
                SCE_CDeleteShaderGLSL(shader->s_glsl[0]);
                Logger_LogSrc();
                return (SCE_ERROR);
            }
        }
        
        /* program */
        shader->p_glsl = SCE_CCreateProgram();
        if(shader->p_glsl == NULL)
        {
            SCE_CDeleteShaderGLSL(shader->s_glsl[1]);
            SCE_CDeleteShaderGLSL(shader->s_glsl[0]);
            Logger_LogSrc();
            return (SCE_ERROR);
        }
        /***/
        
        /* envoi des sources */
        if(shader->vs_addsrc != NULL)
            SCE_CSetShaderGLSLSource(shader->s_glsl[0], shader->vs_addsrc);
        if(shader->vs_source != NULL)
            SCE_CSetShaderGLSLSource(shader->s_glsl[0], shader->vs_source);
        if(shader->ps_addsrc != NULL)
            SCE_CSetShaderGLSLSource(shader->s_glsl[1], shader->ps_addsrc);
        if(shader->ps_source != NULL)
            SCE_CSetShaderGLSLSource(shader->s_glsl[1], shader->ps_source);
        
        /* construction des shaders */
        if(shader->vs_source != NULL || shader->vs_addsrc != NULL)
        {
            /* vertex shader */
            if(SCE_CBuildShaderGLSL(shader->s_glsl[0]) < 0)
            {
                SCE_CDeleteProgram(shader->p_glsl);
                SCE_CDeleteShaderGLSL(shader->s_glsl[1]);
                SCE_CDeleteShaderGLSL(shader->s_glsl[0]);
                Logger_LogSrc();
                return (SCE_ERROR);
            }
        }
        if(shader->ps_source != NULL || shader->ps_addsrc != NULL)
        {
            /* pixel shader */
            if(SCE_CBuildShaderGLSL(shader->s_glsl[1]) < 0)
            {
                SCE_CDeleteProgram(shader->p_glsl);
                SCE_CDeleteShaderGLSL(shader->s_glsl[1]);
                SCE_CDeleteShaderGLSL(shader->s_glsl[0]);
                Logger_LogSrc();
                return (SCE_ERROR);
            }
        }
        /***/
        
        /* association des shaders au program */
        SCE_CSetProgramShader(shader->p_glsl, shader->s_glsl[0], SCE_TRUE);
        if(shader->s_glsl[1] != NULL)
            SCE_CSetProgramShader(shader->p_glsl, shader->s_glsl[1], SCE_TRUE);
        
        /* construction du program GL */
        if(SCE_CBuildProgram(shader->p_glsl) < 0)
        {
            SCE_CDeleteProgram(shader->p_glsl);
            SCE_CDeleteShaderGLSL(shader->s_glsl[1]);
            SCE_CDeleteShaderGLSL(shader->s_glsl[0]);
            Logger_LogSrc();
            return (SCE_ERROR);
        }
    }
    else
    #ifdef _SCE_USE_CG
    if(type == SCE_ASM_SHADER)
    #endif
    {
        /* vertex shader */
        if(shader->vs_source != NULL || shader->vs_addsrc != NULL)
        {
            /* creation de l'objet */
            shader->s_asm[0] = SCE_CCreateShaderASM(SCE_VERTEX_SHADER);
            if(shader->s_asm[0] == NULL)
            {
                Logger_LogSrc();
                return (SCE_ERROR);
            }
            
            /* concatenation des defines avec le code source (vertex) */
            if(shader->vs_addsrc != NULL)
                final_size += strlen(shader->vs_addsrc);
            if(shader->vs_source != NULL)
                final_size += strlen(shader->vs_source);
            final_src = SCE_malloc(final_size);
            if(final_src == NULL)
            {
                SCE_CDeleteShaderASM(shader->s_asm[0]);
                Logger_LogSrc();
                return (SCE_ERROR);
            }
            memset(final_src, '\0', final_size);
            if(shader->vs_addsrc != NULL)
                strcat(final_src, shader->vs_addsrc);
            if(shader->vs_source != NULL)
                strcat(final_src, shader->vs_source);
            SCE_CSetShaderASMSource(shader->s_asm[0], final_src);
            /***/
            
            SCE_free(final_src);
            final_src = NULL;
            final_size = 1;
            
            /* construction du shader */
            if(SCE_CBuildShaderASM(shader->s_asm[0]) < 0)
            {
                SCE_CDeleteShaderASM(shader->s_asm[0]);
                Logger_LogSrc();
                return (SCE_ERROR);
            }
        }
        
        /* pixel shader */
        if(shader->ps_source != NULL || shader->ps_addsrc != NULL)
        {
            /* creation de l'objet */
            shader->s_asm[1] = SCE_CCreateShaderASM(SCE_PIXEL_SHADER);
            if(shader->s_asm[1] == NULL)
            {
                SCE_CDeleteShaderASM(shader->s_asm[0]);
                Logger_LogSrc();
                return (SCE_ERROR);
            }
            
            /* concatenation des defines avec le code source (pixel) */
            if(shader->vs_addsrc != NULL)
                final_size += strlen(shader->vs_addsrc);
            if(shader->vs_source != NULL)
                final_size += strlen(shader->vs_source);
            final_src = SCE_malloc(final_size);
            if(final_src == NULL)
            {
                SCE_CDeleteShaderASM(shader->s_asm[1]);
                SCE_CDeleteShaderASM(shader->s_asm[0]);
                Logger_LogSrc();
                return (SCE_ERROR);
            }
            memset(final_src, '\0', final_size);
            if(shader->vs_addsrc != NULL)
                strcat(final_src, shader->vs_addsrc);
            if(shader->vs_source != NULL)
                strcat(final_src, shader->vs_source);
            SCE_CSetShaderASMSource(shader->s_asm[1], final_src);
            /***/
            
            SCE_free(final_src);
            final_src = NULL;
            final_size = 1;
            
            /* construction du shader */
            if(SCE_CBuildShaderASM(shader->s_asm[1]) < 0)
            {
                SCE_CDeleteShaderASM(shader->s_asm[1]);
                SCE_CDeleteShaderASM(shader->s_asm[0]);
                Logger_LogSrc();
                return (SCE_ERROR);
            }
        }
    }
    #ifdef _SCE_USE_CG
    else /* if(type == SCE_CG_SHADER) */
    {
        /* vertex shader */
        if(shader->vs_source != NULL || shader->vs_addsrc != NULL)
        {
            /* creation de l'objet */
            shader->s_cg[0] = SCE_CCreateShaderCG(SCE_VERTEX_SHADER);
            if(shader->s_cg[0] == NULL)
            {
                Logger_LogSrc();
                return (SCE_ERROR);
            }
            
            /* concatenation des defines avec le code source (vertex) */
            if(shader->vs_addsrc != NULL)
                final_size += strlen(shader->vs_addsrc);
            if(shader->vs_source != NULL)
                final_size += strlen(shader->vs_source);
            final_src = SCE_malloc(final_size);
            if(final_src == NULL)
            {
                SCE_CDeleteShaderCG(shader->s_cg[0]);
                Logger_LogSrc();
                return (SCE_ERROR);
            }
            memset(final_src, '\0', final_size);
            if(shader->vs_addsrc != NULL)
                strcat(final_src, shader->vs_addsrc);
            if(shader->vs_source != NULL)
                strcat(final_src, shader->vs_source);
            SCE_CSetShaderCGSource(shader->s_cg[0], final_src);
            /***/
            
            SCE_free(final_src);
            final_src = NULL;
            final_size = 1;
            
            if(SCE_CBuildShaderCG(shader->s_cg[0]) < 0)
            {
                SCE_CDeleteShaderCG(shader->s_cg[0]);
                Logger_LogSrc();
                return (SCE_ERROR);
            }
        }
        
        /* pixel shader */
        if(shader->ps_source != NULL || shader->ps_addsrc != NULL)
        {
            shader->s_cg[1] = SCE_CCreateShaderCG(SCE_PIXEL_SHADER);
            if(shader->s_cg[1] == NULL)
            {
                SCE_CDeleteShaderCG(shader->s_cg[0]);
                Logger_LogSrc();
                return (SCE_ERROR);
            }
            
            /* concatenation des defines avec le code source (pixel) */
            if(shader->vs_addsrc != NULL)
                final_size += strlen(shader->vs_addsrc);
            if(shader->vs_source != NULL)
                final_size += strlen(shader->vs_source);
            final_src = SCE_malloc(final_size);
            if(final_src == NULL)
            {
                SCE_CDeleteShaderCG(shader->s_cg[1]);
                SCE_CDeleteShaderCG(shader->s_cg[0]);
                Logger_LogSrc();
                return (SCE_ERROR);
            }
            memset(final_src, '\0', final_size);
            if(shader->vs_addsrc != NULL)
                strcat(final_src, shader->vs_addsrc);
            if(shader->vs_source != NULL)
                strcat(final_src, shader->vs_source);
            SCE_CSetShaderCGSource(shader->s_cg[1], final_src);
            /***/
            
            SCE_free(final_src);
            final_src = NULL;
            final_size = 1;
            
            if(SCE_CBuildShaderCG(shader->s_cg[1]) < 0)
            {
                SCE_CDeleteShaderCG(shader->s_cg[0]);
                SCE_CDeleteShaderCG(shader->s_cg[1]);
                Logger_LogSrc();
                return (SCE_ERROR);
            }
        }
    }
    #endif
    
    shader->ready = SCE_TRUE;
    
    return (SCE_OK);
}


int SCE_Shader_AddSource(SCE_SShader *shader, int type, char *src)
{
    /* taille de la reallocation */
    size_t realen;
    
    int first_alloc = 1;
    
    char *addsrc = NULL;
    
    #ifdef __SCE_DEBUG__
    if(shader == NULL || src == NULL)
    {
        Logger_Log(SCE_INVALID_ARG);
        return (SCE_ERROR);
    }
    
    if(type != SCE_PIXEL_SHADER && type != SCE_VERTEX_SHADER &&
       type != SCE_UNKNOW_SHADER)
    {
        Logger_Log(SCE_INVALID_ARG);
        return (SCE_ERROR);
    }
    #endif
    
    
    addsrc =
    (type == SCE_PIXEL_SHADER) ? shader->ps_addsrc:shader->vs_addsrc;
    
    /* + 2 car, 1: retour chariot. 2: caractere de fin de chaine */
    realen = strlen(src) + 2;
    
    if(addsrc != NULL)
    {
        first_alloc = 0;
        realen += strlen(addsrc);
    }
    
    addsrc = SCE_realloc(addsrc, realen);
    if(addsrc == NULL)
    {
        Logger_LogSrc();
        return (SCE_ERROR);
    }
    /* s'il s'agit de la premiere allocation, on initialise le contenu */
    if(first_alloc)
        memset(addsrc, '\0', realen);
    
    strcat(addsrc, src);
    
    /* affectation */
    if(type == SCE_PIXEL_SHADER)
        shader->ps_addsrc = addsrc;
    else
        shader->vs_addsrc = addsrc;
    
    
    return (SCE_OK);
}



/* localise et renvoie un index au sein du code source d'un shader ASM */
static int SCE_Shader_GetASMIndex(const char *src, const char *name)
{
    src = name = NULL;
    return (-1);
}

int SCE_Shader_GetIndex(SCE_SShader *shader, int type, const char *name)
{
    int index;
    
    
    #ifdef __SCE_DEBUG__
    if(shader == NULL || name == NULL)
    {
        Logger_Log(SCE_INVALID_ARG);
        return (-1); /* -1 = mauvais index */
    }
    if(!shader->ready)
    {
        Logger_Log(SCE_INVALID_OPERATION);
        Logger_LogMsg("you can't use a non-builded shader");
        return (-1);
    }
    #endif
    
    if(shader->type == SCE_GLSL_SHADER)
    {
        index = SCE_CGetProgramIndex(shader->p_glsl, name);
    }
    else
    #ifdef _SCE_USE_CG
    if(shader->type == SCE_ASM_SHADER)
    #endif
    {
        /* on va effectuer une judicieuse recherche au sein du code source... */
        if(type == SCE_PIXEL_SHADER)
            index = SCE_Shader_GetASMIndex(shader->ps_source, name);
        else
            index = SCE_Shader_GetASMIndex(shader->vs_source, name);
    }
    #ifdef _SCE_USE_CG
    else
    {
        index = (type == SCE_PIXEL_SHADER) ? 1 : 0;
        
        if(shader->s_cg[index] != NULL)
            index = (int)SCE_CGetShaderCGIndex(shader->s_cg[index], name);
        else
            index = 0;  /* 0 et non -1 : les indices Cg
                           sont des pointeurs sur structure */
    }
    #endif
    
    return (index);
}

/*
int SCE_Shader_GetParam(SCE_SShader*, const char*);
float SCE_Shader_GetParamf(SCE_SShader*, const char*);
float* SCE_Shader_GetParamfv(SCE_SShader*, const char*);
*/

void
SCE_Shader_Param(SCE_SShader *shader, int type, const char *name, int val)
{
    int index;
    
    if(shader != NULL)
    {
        if(shader->type == SCE_GLSL_SHADER)
        {
            SCE_CSetProgramParam(SCE_CGetProgramIndex(shader->p_glsl, name),
                                 val);
        }
        else
        #ifdef _SCE_USE_CG
        if(shader->type == SCE_ASM_SHADER)
        #endif
        {
            if(type == SCE_PIXEL_SHADER)
                index = SCE_Shader_GetASMIndex(shader->ps_source, name);
            else
                index = SCE_Shader_GetASMIndex(shader->vs_source, name);
            
            SCE_CSetShaderASMParamf(index, (float)val, 0.0, 0.0, 1.0);
        }
        #ifdef _SCE_USE_CG
        else
        {
            index = (type == SCE_PIXEL_SHADER) ? 1 : 0;
            
            if(shader->s_cg[index] != NULL)
                index = (int)SCE_CGetShaderCGIndex(shader->s_cg[index], name);
            else
                index = 0;
            
            if(index)
                SCE_CSetShaderCGParam((CGparameter)index, val);
        }
        #endif
    }
}

void
SCE_Shader_Paramf(SCE_SShader *shader, int type, const char *name, float val)
{
    int index;
    
    if(shader != NULL)
    {
        if(shader->type == SCE_GLSL_SHADER)
        {
            SCE_CSetProgramParamf(SCE_CGetProgramIndex(shader->p_glsl, name),
                                  val);
        }
        else
        #ifdef _SCE_USE_CG
        if(shader->type == SCE_ASM_SHADER)
        #endif
        {
            if(type == SCE_PIXEL_SHADER)
                index = SCE_Shader_GetASMIndex(shader->ps_source, name);
            else
                index = SCE_Shader_GetASMIndex(shader->vs_source, name);
            
            SCE_CSetShaderASMParamf(index, val, 0.0, 0.0, 1.0);
        }
        #ifdef _SCE_USE_CG
        else
        {
            index = (type == SCE_PIXEL_SHADER) ? 1 : 0;
            
            if(shader->s_cg[index] != NULL)
                index = (int)SCE_CGetShaderCGIndex(shader->s_cg[index], name);
            else
                index = 0;
            
            if(index)
                SCE_CSetShaderCGParamf((CGparameter)index, val);
        }
        #endif
    }
}

void
SCE_Shader_Paramfv(SCE_SShader *shader, int type, const char *name,
                      size_t size, float *val)
{
    int index;
    
    if(shader != NULL)
    {
        if(shader->type == SCE_GLSL_SHADER)
        {
            SCE_CSetProgramParamfv(SCE_CGetProgramIndex(shader->p_glsl, name),
                                   size,
                                   val);
        }
        else
        #ifdef _SCE_USE_CG
        if(shader->type == SCE_ASM_SHADER)
        #endif
        {
            if(type == SCE_PIXEL_SHADER)
                index = SCE_Shader_GetASMIndex(shader->ps_source, name);
            else
                index = SCE_Shader_GetASMIndex(shader->vs_source, name);
            
            SCE_CSetShaderASMParamfv(index, size, val);
        }
        #ifdef _SCE_USE_CG
        else
        {
            index = (type == SCE_PIXEL_SHADER) ? 1 : 0;
            
            if(shader->s_cg[index] != NULL)
                index = (int)SCE_CGetShaderCGIndex(shader->s_cg[index], name);
            else
                index = 0;
            
            if(index)
                SCE_CSetShaderCGParamfv((CGparameter)index, size, val);
        }
        #endif
    }
}


void SCE_Shader_SetParam(SCE_SShader *shader, int index, int val)
{
    if(shader != NULL)
    {
        if(shader->type == SCE_GLSL_SHADER)
        {
            SCE_CSetProgramParam(index, val);
        }
        else
        #ifdef _SCE_USE_CG
        if(shader->type == SCE_ASM_SHADER)
        #endif
        {
            SCE_CSetShaderASMParamf(index, val, 0.0, 0.0, 1.0);
        }
        #ifdef _SCE_USE_CG
        else
        {
            if(index)
                SCE_CSetShaderCGParam((CGparameter)index, val);
        }
        #endif
    }
}

void SCE_Shader_SetParamf(SCE_SShader *shader, int index, float val)
{
    if(shader != NULL)
    {
        if(shader->type == SCE_GLSL_SHADER)
        {
            SCE_CSetProgramParamf(index, val);
        }
        else
        #ifdef _SCE_USE_CG
        if(shader->type == SCE_ASM_SHADER)
        #endif
        {
            SCE_CSetShaderASMParamf(index, val, 0.0, 0.0, 1.0);
        }
        #ifdef _SCE_USE_CG
        else
        {
            if(index)
                SCE_CSetShaderCGParamf((CGparameter)index, val);
        }
        #endif
    }
}

void
SCE_Shader_SetParamfv(SCE_SShader *shader, int index, size_t size, float *val)
{
    if(shader != NULL)
    {
        if(shader->type == SCE_GLSL_SHADER)
        {
            SCE_CSetProgramParamfv(index, size, val);
        }
        else
        #ifdef _SCE_USE_CG
        if(shader->type == SCE_ASM_SHADER)
        #endif
        {
            SCE_CSetShaderASMParamfv(index, size, val);
        }
        #ifdef _SCE_USE_CG
        else
        {
            if(index)
                SCE_CSetShaderCGParamfv((CGparameter)index, size, val);
        }
        #endif
    }
}



void SCE_Shader_Use(SCE_SShader *shader)
{
    static int back_language = SCE_UNKNOW_SHADER;
    static void *pixelshader = NULL;
    static int vs_binded = SCE_FALSE;
    
    if(shader != NULL)
    {
        if(shader->type == SCE_GLSL_SHADER)
        {
            if(back_language == SCE_ASM_SHADER)
            {
                SCE_CUseShaderASM(NULL);    /* vertex shader */
                SCE_CUseShaderASM(NULL);    /* pixel shader */
            }
            #ifdef _SCE_USE_CG
            else if(back_language == SCE_CG_SHADER)
            {
                SCE_CUseShaderCG(NULL);     /* vertex shader */
                SCE_CUseShaderCG(NULL);     /* pixel shader */
            }
            #endif
            
            SCE_CUseProgram(shader->p_glsl);
            
            pixelshader = NULL;
        }
        else
        #ifdef _SCE_USE_CG
        if(shader->type == SCE_ASM_SHADER)
        #endif
        {
            /* desactivation des shaders precedement actives */
            if(back_language == SCE_GLSL_SHADER)
            {
                SCE_CUseProgram(NULL);
                vs_binded = SCE_FALSE;
            }
            #ifdef _SCE_USE_CG
            else if(back_language == SCE_CG_SHADER)
            {
                SCE_CUseShaderCG(NULL);     /* vertex shader */
                SCE_CUseShaderCG(NULL);     /* pixel shader */
                vs_binded = SCE_FALSE;
            }
            #endif
            
            /* analyse du type du shader */
            if(shader->s_asm[0] == NULL)
            {
                /* aucun vertex shader n'est present, on verifie s'il y en
                   a deja un d'actif, si oui, on active le pixel shader
                   du shader envoye, sinon, on retient l'adresse du pixel shader
                   pour une activation ulterieur */
                if(vs_binded)
                    SCE_CUseShaderASM(shader->s_asm[1]);
                else
                    pixelshader = shader->s_asm[1];
            }
            else
            {
                /* un vertex shader est present, on l'active */
                SCE_CUseShaderASM(shader->s_asm[0]);
                
                /* si un pixel shader est present, on l'active */
                if(shader->s_asm[1])
                    SCE_CUseShaderASM(shader->s_asm[1]);
                /* sinon, si un pixel shader de sauvegarde est present,
                   on active celui-ci */
                else if(pixelshader)
                {
                    SCE_CUseShaderASM(pixelshader);
                    pixelshader = NULL;
                }
                
                /* un vertex shader est actif */
                vs_binded = SCE_TRUE;
            }
        }
        #ifdef _SCE_USE_CG
        else
        {
            if(back_language == SCE_GLSL_SHADER)
            {
                SCE_CUseProgram(NULL);
                vs_binded = SCE_FALSE;
            }
            else if(back_language == SCE_ASM_SHADER)
            {
                SCE_CUseShaderASM(NULL);    /* vertex shader */
                SCE_CUseShaderASM(NULL);    /* pixel shader */
                vs_binded = SCE_FALSE;
            }
            
            if(shader->s_cg[0] == NULL)
            {
                if(vs_binded)
                    SCE_CUseShaderCG(shader->s_cg[1]);
                else
                    pixelshader = shader->s_cg[1];
            }
            else
            {
                SCE_CUseShaderCG(shader->s_cg[0]);
                if(shader->s_cg[1])
                    SCE_CUseShaderCG(shader->s_cg[1]);
                else if(pixelshader)
                {
                    SCE_CUseShaderCG(pixelshader);
                    pixelshader = NULL;
                }
                
                vs_binded = SCE_TRUE;
            }
        }
        #endif
        
        back_language = shader->type;
    }
    else
    {
        if(back_language == SCE_GLSL_SHADER)
            SCE_CUseProgram(NULL);
        else if(back_language == SCE_ASM_SHADER)
        {
            SCE_CUseShaderASM(NULL);    /* vertex shader */
            SCE_CUseShaderASM(NULL);    /* pixel shader */
        }
        #ifdef _SCE_USE_CG
        else if(back_language == SCE_CG_SHADER)
        {
            SCE_CUseShaderCG(NULL);     /* vertex shader */
            SCE_CUseShaderCG(NULL);     /* pixel shader */
        }
        #endif
        
        pixelshader = NULL;
        vs_binded = SCE_FALSE;
        back_language = SCE_UNKNOW_SHADER;
    }
}
